package ch.dvbern.stvbe.security.saml.handler;

import java.util.Collections;
import java.util.Map;
import java.util.Set;

import javax.xml.namespace.QName;
import javax.xml.ws.handler.MessageContext;
import javax.xml.ws.handler.MessageContext.Scope;
import javax.xml.ws.handler.soap.SOAPHandler;
import javax.xml.ws.handler.soap.SOAPMessageContext;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import ch.dvbern.stvbe.security.saml.handler.extract.SamlAssertionExtractor;
import ch.dvbern.stvbe.security.saml.handler.extract.SoapMsgAssertionExtractor;
import ch.dvbern.stvbe.security.saml.handler.util.SamlValidationException;
import ch.dvbern.stvbe.security.saml.handler.util.SamlValidationResult;
import ch.dvbern.stvbe.security.saml.handler.util.SamlValidationUtil;
import ch.dvbern.stvbe.security.saml.verify.SamlAssertionVerifier;
import ch.dvbern.stvbe.security.saml.verify.SamlAssertionVerifierFactory;

/**
 * SAML Validierungs-Handler fuer JAX-WS Web Services
 * @author GEMA
 */
public class JaxWsSamlValidationHandler implements SOAPHandler<SOAPMessageContext>{
	/**
	 * Name des Message Context Properties fuer den Principal Namen. Der Handler Legt den Principal
	 * Namen unter diesem Key im {@link MessageContext} ab.
	 * @see MessageContext
	 */
	public static final String PRINCIPAL_NAME_PROP = HandlerConstants.PRINCIPAL_NAME_PROP;

	/**
	 * Name des Message Context Properties fuer die SAML Assertion. Der Handler Legt die SAML Assertion
	 * unter diesem Key im {@link MessageContext} ab.
	 * @see MessageContext
	 */
	public static final String ASSERTION_PROP = HandlerConstants.ASSERTION_PROP;
	
	private static final Log log = LogFactory.getLog(JaxWsSamlValidationHandler.class);

	private static final Set<QName> HEADERS = Collections.singleton(HandlerConstants.SECURITY_HEADER_QNAME);
	
	private SamlAssertionVerifier verifier;
	
	/**
	 * Konstruktor
	 */
	public JaxWsSamlValidationHandler() {
	}

	public Set<QName> getHeaders() {
		return HEADERS;
	}

	/**
	 * Initialisiert den Handler
	 * @param config die Handler Konfiguration
	 */
	public void init(Map<String,String> config) {
		try {
			verifier = SamlAssertionVerifierFactory.getInstance().createVerifier(config);
			log.debug("SAML Handler configured. running in native mode.");
		} catch (RuntimeException e) {
			log.error("Handler initialization failed", e);
		}		
	}
	
	public boolean handleMessage(SOAPMessageContext msgCtx) {
		if (isOutboundMessage(msgCtx)) {
			return true;
		} else {
			return handleRequest(msgCtx);
		}
	}
	
	private boolean handleRequest(SOAPMessageContext msgCtx) {
		log.trace("Starting SAML Valdiation");

		if (verifier == null) {
			throw new IllegalStateException("configuration not set");
		}
		
		try {
			
			SamlAssertionExtractor provider = getAssertionProvider(msgCtx);

			SamlValidationResult result =  SamlValidationUtil.validate(provider, verifier);

			// Properties setzen
			msgCtx.put(ASSERTION_PROP, result.getAssertion());
			msgCtx.put(PRINCIPAL_NAME_PROP, result.getPrincipalName());

			// Sichtbarkeit der Properties bestimmen
			msgCtx.setScope(ASSERTION_PROP, Scope.APPLICATION);
			msgCtx.setScope(PRINCIPAL_NAME_PROP, Scope.APPLICATION);
			
			log.trace("SAML Valdiation finished");
			return true;
		} catch (RuntimeException e) {
			log.info("SAML Validation failed", e);
			throw e;
		} catch (SamlValidationException e) {
			log.error("SAML Validation failed", e);
			throw new RuntimeException("SAML validation failed", e);
		} 
	}

	/**
	 * Erstellt den Assertion Provider
	 * @param context der Kontext
	 * @return den Provider
	 */
	@SuppressWarnings("unchecked")
	protected SamlAssertionExtractor getAssertionProvider(SOAPMessageContext context) {
		return new SoapMsgAssertionExtractor(context.getMessage());
	}


	public boolean handleFault(SOAPMessageContext arg0) {
		return true;
	}

	public void close(MessageContext arg0) {
	}

	/**
	 * Gibt zurueck, ob die uebergebene Meldung eine ausgehende Meldung ist
	 * @param ctx der Context
	 * @return <code>true</code> falls die Meldung ausgehend ist
	 */
	private boolean isOutboundMessage(MessageContext ctx) {
		Boolean b = (Boolean)ctx.get(MessageContext.MESSAGE_OUTBOUND_PROPERTY);
		if (b == null) {
			return false;
		} else {
			return b.booleanValue();
		}
	}
}
