package ch.dvbern.stvbe.security.saml.handler;

import java.util.Map;
import java.util.Vector;

import javax.xml.namespace.QName;
import javax.xml.rpc.JAXRPCException;
import javax.xml.rpc.handler.GenericHandler;
import javax.xml.rpc.handler.HandlerInfo;
import javax.xml.rpc.handler.MessageContext;
import javax.xml.rpc.handler.soap.SOAPMessageContext;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.ws.security.handler.WSHandlerConstants;
import org.apache.ws.security.handler.WSHandlerResult;

import ch.dvbern.stvbe.security.saml.handler.extract.SamlAssertionExtractor;
import ch.dvbern.stvbe.security.saml.handler.extract.SoapMsgAssertionExtractor;
import ch.dvbern.stvbe.security.saml.handler.extract.Wss4jAssertionExtractor;
import ch.dvbern.stvbe.security.saml.handler.util.SamlValidationException;
import ch.dvbern.stvbe.security.saml.handler.util.SamlValidationResult;
import ch.dvbern.stvbe.security.saml.handler.util.SamlValidationUtil;
import ch.dvbern.stvbe.security.saml.util.ConfigKeys;
import ch.dvbern.stvbe.security.saml.verify.SamlAssertionVerifier;
import ch.dvbern.stvbe.security.saml.verify.SamlAssertionVerifierFactory;

/**
 * SAML Validierungs-Handler fuer JAX-RPC Web Services (z.B. Axis 1.x)
 * @author GEMA
 */
public class JaxrpcSamlValidationHandler extends GenericHandler {

	/**
	 * Name des Message Context Properties fuer den Principal Namen. Der Handler Legt den Principal
	 * Namen unter diesem Key im {@link MessageContext} ab.
	 * @see MessageContext
	 */
	public static final String PRINCIPAL_NAME_PROP = HandlerConstants.PRINCIPAL_NAME_PROP;

	/**
	 * Name des Message Context Properties fuer die SAML Assertion. Der Handler Legt die SAML Assertion
	 * unter diesem Key im {@link MessageContext} ab.
	 * @see MessageContext
	 */
	public static final String ASSERTION_PROP = HandlerConstants.ASSERTION_PROP;
	
	private static final Log log = LogFactory.getLog(JaxrpcSamlValidationHandler.class);

	private static final QName[] HEADERS = new QName[] { HandlerConstants.SECURITY_HEADER_QNAME };
	
	private SamlAssertionVerifier verifier;
	
	private boolean wss4jBridgeMode;

	/**
	 * Konstruktor
	 */
	public JaxrpcSamlValidationHandler() {
	}


	public QName[] getHeaders() {
		return HEADERS;
	}
	
	@SuppressWarnings("unchecked")
	@Override
	public void init(HandlerInfo handlerInfo) {
		try {
			super.init(handlerInfo);
			Map<String, String> config = handlerInfo.getHandlerConfig();
			verifier = SamlAssertionVerifierFactory.getInstance()
					.createVerifier(config);
			wss4jBridgeMode = Boolean.parseBoolean(config
					.get(ConfigKeys.WSS4J_BRIDGE));
			if (wss4jBridgeMode) {
				log
						.debug("SAML Handler configured. running in WSS4J bridge mode.");
			} else {
				log.debug("SAML Handler configured. running in native mode.");
			}
		} catch (RuntimeException e) {
			log.error("Handler initialization failed", e);
		}		
	}
	
	
	public boolean handleRequest(MessageContext msgCtx) {
		log.trace("Starting SAML Valdiation");

		if (verifier == null) {
			throw new IllegalStateException("configuration not set");
		}
		
		try {
			SOAPMessageContext soapCtx = (SOAPMessageContext) msgCtx;
			SamlAssertionExtractor provider = getAssertionProvider(soapCtx);

			SamlValidationResult result =  SamlValidationUtil.validate(provider, verifier);
			
			msgCtx.setProperty(ASSERTION_PROP, result.getAssertion());
			msgCtx.setProperty(PRINCIPAL_NAME_PROP, result.getPrincipalName());
			
			log.trace("SAML Valdiation finished");
			return true;
		} catch (RuntimeException e) {
			log.info("SAML Validation failed", e);
			throw e;
		} catch (SamlValidationException e) {
			log.error("SAML Validation failed", e);
			throw new JAXRPCException("SAML validation failed", e);
		} 
	}

	/**
	 * Erstellt den Assertion Provider
	 * @param context der Kontext
	 * @return den Provider
	 */
	@SuppressWarnings("unchecked")
	protected SamlAssertionExtractor getAssertionProvider(SOAPMessageContext context) {
		
		if (wss4jBridgeMode) {
			Vector<WSHandlerResult> result = (Vector<WSHandlerResult>) context
					.getProperty(WSHandlerConstants.RECV_RESULTS);
			return new Wss4jAssertionExtractor(result);
		} else {
			return new SoapMsgAssertionExtractor(context.getMessage());
		}
	}
	

}
