package ch.dvbern.stvbe.security.saml.handler.extract;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.xml.soap.Name;
import javax.xml.soap.SOAPElement;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPHeader;
import javax.xml.soap.SOAPMessage;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.dom.DOMSource;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.opensaml.SAMLAssertion;
import org.opensaml.SAMLException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import ch.dvbern.stvbe.security.saml.util.NsConstants;
import ch.dvbern.stvbe.security.saml.util.XPathHelper;
import ch.dvbern.stvbe.security.saml.verify.SamlAssertionVerifyerException;

/**
 * Extrahiert SAML Assertions aus einer SOAP Message
 * @author GEMA
 */
public class SoapMsgAssertionExtractor implements SamlAssertionExtractor {

	private static final Log log = LogFactory.getLog(SoapMsgAssertionExtractor.class);

	private static final String WSS_SECURITY_ELEM = "Security";

	private static final String SAML_NS = NsConstants.SAML;

	private static final String SAML_ASSERTION_ELEM = "Assertion";

	private static final String SAML_ASSERTION_QUERY = "//saml:Assertion";

	private SOAPMessage msg;

	private SAMLAssertion assertion;
	
	/**
	 * Konstruktor
	 * @param msg die SOAP Meldung
	 */
	public SoapMsgAssertionExtractor(SOAPMessage msg) {
		this.msg = msg;
	}
	
	public SAMLAssertion getSAMLAssertion() {
		if (assertion == null) {
			processMessage();
		} 
		return assertion;
	}

	/**
	 * Behandelt einen SOAP Request
	 * 
	 * @param msg
	 *            die SOAP Meldung
	 * @throws SamlAssertionVerifyerException
	 *             falls die SAML Validierung fehlschl�gt
	 */
	@SuppressWarnings("unchecked")
	private void processMessage() {

		// nichts tun, falls kein header gefunden wird
		try {
			SOAPHeader header = msg.getSOAPPart().getEnvelope().getHeader();
			if (header == null) {
				return;
			}

			List<SOAPElement> toRemove = new ArrayList<SOAPElement>();

			boolean found = false;

			for (Iterator iter = header.getChildElements(); iter.hasNext();) {
				Object obj = iter.next();
				if (obj instanceof SOAPElement) {
					SOAPElement element = (SOAPElement) obj;
	
					// Saml-Assertion suchen
					Name name = element.getElementName();
					if (NsConstants.WSS.equals(name.getURI())
							&& WSS_SECURITY_ELEM.equals(name.getLocalName())) {
						log.debug("WSS Security Header found");
						found = true;
						assertion = handleSecurityHeader(msg, element);
						toRemove.add(element);
					}
				}
			}

			// Security Headers, welche erfoglreich bearbeitet wurden, entfernen
			for (SOAPElement elem : toRemove) {
				elem.detachNode();
				
			}
			if (!found && log.isDebugEnabled()) {
				log.debug("No wss:Security Header has been found!");
			}
		} catch (SOAPException e) {
			log.error("SAML Validation failed", e);
			throw new SamlAssertionExtractorException("SAML Validation failed", e);
		} 

	}

	/**
	 * Behandelt einen Security Header
	 * 
	 * @param securityHeader das Element
	 */
	@SuppressWarnings("unchecked")
	private SAMLAssertion handleSecurityHeader(SOAPMessage msg, SOAPElement securityHeader) {

		for (Iterator iter = securityHeader.getChildElements(); iter.hasNext();) {
			SOAPElement element = (SOAPElement) iter.next();

			Name name = element.getElementName();
			// SAML Assertions suchen
			if (SAML_NS.equals(name.getURI())
					&& SAML_ASSERTION_ELEM.equals(name.getLocalName())) {
				log.trace("SAML Assertion found");

				return getSAMLAssertion(element);
			}
		}

		return null;
	}
	
	/**
	 * Extrahiert eine SAML Assertion aus einem entsprechenden Element
	 * @param element das Element
	 * @return die Assertion
	 */
	@SuppressWarnings("unchecked")
	private SAMLAssertion getSAMLAssertion(SOAPElement element) {
		// Eine AXIS Message muss zuerst in ein DOM Dokument
		// konvertiert werden. Andernfalls ist die Signatur 
		// des SAML Tokens nicht gueltig
		
		Element domElement = null;
		
		try {
			// Klasse dynamisch laden, damit sie nicht zwingend im
			// Klassenpfad sein muss
			Class axisMsgClass = Class.forName("org.apache.axis.Message");
			if (axisMsgClass.isInstance(msg)) {
				log.debug("AXIS Message Found");
				try {
					// Message in ein Dokument konvertieren
					Document doc = ((org.apache.axis.Message) msg).getSOAPEnvelope()
							.getAsDocument();

					// SAML Assertion mittels XPath Expression suchen
					XPath xPath = XPathHelper.createXPath();
					domElement = (Element) xPath.evaluate(
							SAML_ASSERTION_QUERY, doc,
							XPathConstants.NODE);
				} catch (Exception e) {
					throw new SamlAssertionExtractorException(
							"Axis Envelope could not be converted to a document",
							e);
				}
			}
		} catch (ClassNotFoundException e1) {
			log.debug("AXIS not installed");

			// SAML Assertion Element mit einer Transformation extrahieren
			try {
				DOMSource ds = new DOMSource((Element) element);
				DOMResult dr = new DOMResult();
				TransformerFactory.newInstance().newTransformer().transform(ds, dr);
				domElement = ((Document)dr.getNode()).getDocumentElement();
			} catch (TransformerConfigurationException e) {
				throw new SamlAssertionExtractorException("Can't extract SAML Assertion Element", e);
			} catch (TransformerException e) {
				throw new SamlAssertionExtractorException("Can't extract SAML Assertion Element", e);
			} catch (TransformerFactoryConfigurationError e) {
				throw new SamlAssertionExtractorException("Can't extract SAML Assertion Element", e);
			}
		}
		
		// Login mit der SAML Assertion durchfuehren
		try {
			return new SAMLAssertion(domElement);
		} catch (SAMLException e) {
			throw new SamlAssertionExtractorException("Invalid SAML Assertion", e);
		}		
	}
	
}
