package ch.dvbern.stvbe.security.saml.key;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.security.GeneralSecurityException;
import java.security.KeyStore;
import java.security.PublicKey;
import java.security.cert.Certificate;
import java.security.cert.X509Certificate;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * Implementation eines PublicKeyProviders, welcher die {@link PublicKey}s aus
 * {@link KeyStore} laedt
 * 
 * @author gema
 * 
 */
public class KeyStoreX509CertificatePublicKeyProvider implements PublicKeyProvider {

    private static final Log log = LogFactory.getLog(KeyStoreX509CertificatePublicKeyProvider.class);
    
	private String keyStorePath;

    private String keyStorePassword;

	/**
	 * Konstruktor
	 */
	public KeyStoreX509CertificatePublicKeyProvider() {
	}
	
	/**
	 * Konstruktor
	 * @param keyStorePath der Pfad zum KeyStore
	 * @param keyStorePassword das Passwort fuer den Keystore
	 */
	public KeyStoreX509CertificatePublicKeyProvider(String keyStorePath, String keyStorePassword) {
		if (keyStorePath == null || keyStorePath.trim().length() == 0) {
			throw new IllegalArgumentException("certificate resource name may not be empty");
		}
		this.keyStorePath = keyStorePath;
		this.keyStorePassword = keyStorePassword;
	}
	
	/**
	 * Setzt den Pfad zum KeyStore
	 * @param path der Pfad
	 */
	public void setKeyStorePath(String path) {
		this.keyStorePath = path;
	}
	
    /**
     * Setzt das Passwort fuer den  KeyStore
     * @param password das Passwort
     */
    public void setKeyStorePassword(String password) {
        this.keyStorePassword = password;
    }

    public Set<PublicKey> getPublicKeys() {
		if (keyStorePath == null) {
			throw new PublicKeyProviderException("keyStorePath not set");
		}
		
		String resourceName = keyStorePath;
		if (!resourceName.startsWith("/")) {
		    resourceName = "/" + resourceName;
		}
		
		// 1. aus Classpath laden
		InputStream input = getClass().getResourceAsStream(resourceName);
		if (input == null) {
		    log.trace("Cannot find resource: "+ resourceName + " in classpath. trying in File System.");

		    // 2. aus Filesystem laden
	        try {
	            input = new BufferedInputStream(new FileInputStream(keyStorePath));
	        } catch (FileNotFoundException e1) {
			    log.trace("Cannot find file "+ keyStorePath + " in filesystem. trying as URI.");
	        	
	        	
	        	// 3. Keystore-Pfad als URL interpretieren
				try {
					input = new BufferedInputStream(new FileInputStream(
							new File(new URI(keyStorePath))));
				} catch (Exception e2) {
					throw new PublicKeyProviderException(
							"Cannot find KeyStore File " + keyStorePath
									+ " in classpath or file system", e2);
				}
	        }	
		}
		
        // Zertifikate aus dem Keystore Laden
        Set<PublicKey> out = new HashSet<PublicKey>();
        try {

            KeyStore trustStore = KeyStore.getInstance(KeyStore.getDefaultType());
            trustStore.load(input, keyStorePassword.toCharArray());
            
            for (Enumeration<String> aliases = trustStore.aliases(); aliases.hasMoreElements();) {
                String alias = aliases.nextElement();
                Certificate cert = trustStore.getCertificate(alias);
                if (cert instanceof X509Certificate) {
                    X509Certificate x509Cert = (X509Certificate) cert;
                    out.add(x509Cert.getPublicKey());
                }
            }
            
            if (out.isEmpty()) {
                throw new PublicKeyProviderException("no public key found in keystore " + keyStorePath);
            }
            return out;
        } catch (IOException e) {
            throw new PublicKeyProviderException("can't load keystore", e);
        } catch (GeneralSecurityException e) {
            throw new PublicKeyProviderException("can't load keystore", e);
        } finally {
            try {
                input.close();
            } catch (IOException e) {
                log.warn("can't close stream", e);
            }
        }
	}
}
