package ch.dvbern.stvbe.security.saml.simplewscall;

import java.io.ByteArrayInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.net.HttpURLConnection;
import java.net.URI;
import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;

import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSocketFactory;
import javax.net.ssl.X509TrustManager;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import ch.dvbern.stvbe.security.saml.util.NamespaceContextImpl;
import ch.dvbern.stvbe.security.saml.util.StringUtils;
import ch.dvbern.stvbe.security.saml.util.Validate;



/**
 * Einfache Klasse um einen Webservice basierend auf ein Template-Request aufzurufen. <br>
 * Wird verwendet um möglichst wenig externe Abhängigkeiten zu haben.
 * <p>
 * Diese Klasse ist nicht thread-safe, sie darf nur von einem Thread verwendet werden.
 * 
 * @author ea38
 */
public class SimpleWebserviceCall
{
	private static SSLSocketFactory trustAllSocketFactory;

    private enum State
    {
        /**
         * Call ist vorbereitet 
         */
        PREPARE, 
        /**
         * Call wurde ausgefuehrt
         */
        EXECUTED
    }

    private final URI mUri;
    private State mState;

    private String mRequest;

    private Document mResponse;
    private String mResponseString;
    private NamespaceContextImpl mResponseNamespaceContext = new NamespaceContextImpl();
    {
        mResponseNamespaceContext.bindPrefixToNamespaceURI("soap11", "http://schemas.xmlsoap.org/soap/envelope/");
        mResponseNamespaceContext.bindPrefixToNamespaceURI("soap12", "http://www.w3.org/2003/05/soap-envelope");
    }
    
    private boolean sslCertifcateValidationDisabled = false;
    /**
     * Legt fest, ob die SSL Zertifikatsvalidierung ausgeschaltet ist
     * @param sslCertifcateValidationDisabled <code>true</code> zum Ausschalten der Validierung
     */
    public void setSslCertifcateValidationDisabled(boolean sslCertifcateValidationDisabled) {
		this.sslCertifcateValidationDisabled = sslCertifcateValidationDisabled;
	}
    /**
     * Gibt zurck, ob die SSL Zertifikatsvalidierung ausgeschaltet ist
     * @return <code>true</code> falls die Validierung ausgeschaltet ist
     */
    public boolean isSslCertifcateValidationDisabled() {
		return sslCertifcateValidationDisabled;
	}
    
    /**
     * Neuen Webservice-Call erstellen.
     * 
     * @param webserviceURI
     *            URI des aufzurufenden Webservices, nicht <code>null</code>
     * @param caller
     *            Aufrufende Klasse, wird verwendet um relative Ressourcen zu laden, nicht
     *            <code>null</code>
     * @param requestTemplate
     *            Name des Templates, relativ zu caller, nicht <code>null</code>
     * @throws RequestTemplateNotFoundException
     */
    public SimpleWebserviceCall(URI webserviceURI, Class<?> caller, String requestTemplate)
        throws RequestTemplateNotFoundException
    {
        Validate.notNull(webserviceURI);
        Validate.notNull(caller);
        Validate.notNull(requestTemplate);

        mUri = webserviceURI;
        mRequest = readRequest(caller, requestTemplate);
        mState = State.PREPARE;
    }

    /**
     * Prueft, ob der Status korrekt ist
     * @param should der erwartete Status
     * @throws IllegalStateException falls der aktuelle Status nicht gleich wie der erwartete ist
     */
    protected void assertState(State should)
    {
        if (!should.equals(mState))
            throw new IllegalStateException("Wrong state " + mState + " instead of " + should);
    }

    /**
     * Setzt einen Parameter fuer den Aufruf.<br>
     * Parameter werden im Template-Request mit ${&lt;name&gt;} gekennzeichnet (z. B. ${version})
     * 
     * @param name
     *            nicht <code>null</code>
     * @param value
     *            <code>null</code> entspricht einem leeren String
     */
    public void setParameter(String name, String value)
    {
        assertState(State.PREPARE);
        Validate.notNull(name);
        if (value == null)
            value = "";

        String replace = "${" + name + "}";
        mRequest = StringUtils.replace(mRequest, replace, value);
    }

    /**
     * 
     * @return den Request
     */
    protected String getRequest()
    {
        return mRequest;
    }

    /**
     * Führt den Request durch.
     * 
     * @return die Response als geparstes DOM-Document, nicht <code>null</code> (das gleiche wie
     *         {@link #getResponse()})
     */
    public Document execute()
    {
        assertState(State.PREPARE);

        try
        {
            HttpURLConnection connection = (HttpURLConnection) mUri.toURL().openConnection();
    		if (connection instanceof HttpsURLConnection && sslCertifcateValidationDisabled) {
    			HttpsURLConnection httpsCon = (HttpsURLConnection) connection;
    			httpsCon.setSSLSocketFactory(getTrustAllSocketFactory());
    		}            

            connection.setDoOutput(true);
            connection.setDoInput(true);
            

            OutputStream os = connection.getOutputStream();
            os.write(getRequest().getBytes("UTF-8"));
            connection.connect();

            // Bei Errors wirft getInputStream einen Fehler. Bei Error-Code 50 wollen wir aber die
            InputStream is;
            if (connection.getResponseCode() != HttpURLConnection.HTTP_INTERNAL_ERROR)
                is = connection.getInputStream();
            else
                is = connection.getErrorStream();

            try
            {
                String responseString = readStreamToString(is);

                mResponse = parseToDom(new ByteArrayInputStream(responseString.getBytes("UTF-8")));
                mResponseString = responseString;

                mState = State.EXECUTED;
                return mResponse;
            }
            finally
            {
                saveClose(is);
            }
        }
        catch (ClassCastException e)
        {
            throw new WebserviceCallException("Unsupported URL, has to be HTTP(S)");
        }
        catch (FileNotFoundException e)
        {
            throw new WebserviceCallException("URL " + mUri
                + " is possibly wrong, the URL was not found: " + e.getMessage());
        }
        catch (IOException e)
        {
            throw new WebserviceCallException(e);
        }
    }

    /**
     * Die Antwort des Webservices als DOM-Document.
     * 
     * @return nicht <code>null</code>
     */
    public Document getResponse()
    {
        assertState(State.EXECUTED);
        return mResponse;
    }

    /**
     * @return Antwort als String
     */
    public String getResponseAsString()
    {
        assertState(State.EXECUTED);
        return mResponseString;
    }

    /**
     * Document-Element der Anwort.
     * 
     * @return nicht <code>null</code>
     */
    public Element getResponseElement()
    {
        return getResponse().getDocumentElement();
    }

    /**
     * Fügt einen Namespace hinzu, der zum Parsen der Response verwendet werden kann (XPath).
     * 
     * @param prefix
     *            nicht <code>null</code>
     * @param uri
     *            nicht <code>null</code>
     */
    public void addNamespace(String prefix, String uri)
    {
        mResponseNamespaceContext.bindPrefixToNamespaceURI(prefix, uri);
    }

    /**
     * XPath-Ausdruck auf die Antwort des WS ausführen.
     * 
     * @see XPath#evaluate(String, Object)
     * @param expression
     *            nicht <code>null</code>
     * @return kann <code>null</code> sein
     * @throws XPathExpressionException
     */
    public String evalutateXPath(String expression) throws XPathExpressionException
    {
        Validate.notNull(expression);
        return getXPath().evaluate(expression, getResponseElement());
    }

    /**
     * Ein bestimmmtes Element in der Response zurückgeben.
     * 
     * @param elementName
     *            nicht <code>null</code>
     * @return <code>null</code>
     */
    public Element getElementByTagName(String elementName)
    {
        Validate.notNull(elementName);
        NodeList elements = getResponseElement().getElementsByTagName("Assertion");
        if (elements.getLength() == 0)
            return null;
        return (Element) elements.item(0);
    }

    /**
     * @return <code>true</code> wenn die Reponse ein SOAP-Fault ist
     * @throws XPathExpressionException
     */
    public boolean isResponseFault() throws XPathExpressionException
    {
        Node node = (Node) getXPath().evaluate("/soap11:Envelope/soap11:Body/soap11:Fault",
            getResponseElement(), XPathConstants.NODE);
        if (node == null)
        {
            node = (Node) getXPath().evaluate("/soap12:Envelope/soap12:Body/soap12:Fault",
                getResponseElement(), XPathConstants.NODE);
        }
        return node != null;
    }

    /**
     * Die Fehlermeldung im SOAP-Fault (faultstring).
     * 
     * @return <code>null</code> wenn kein Fault oder keinen faultstring in der Fehlermeldung
     * @throws XPathExpressionException
     */
    public String getFaultMessage() throws XPathExpressionException
    {
        String faultstring = getXPath().evaluate(
            "/soap11:Envelope/soap11:Body/soap11:Fault/soap11:faultstring", getResponseElement());
        if (StringUtils.isEmpty(faultstring))
        {
            faultstring = getXPath().evaluate("/soap11:Envelope/soap11:Body/soap11:Fault/faultstring",
                getResponseElement());
        }
        if (StringUtils.isEmpty(faultstring))
        {
            faultstring = getXPath().evaluate(
                "/soap12:Envelope/soap12:Body/soap12:Fault/soap12:faultstring", getResponseElement());
        }
        if (StringUtils.isEmpty(faultstring))
        {
            faultstring = getXPath().evaluate("/soap12:Envelope/soap12:Body/soap12:Fault/faultstring",
                getResponseElement());
        }

        if (StringUtils.isEmpty(faultstring))
            return null;
        return faultstring;
    }

    /**
     * Die Fehlermeldung im SOAP-Fault (faultstring).
     * 
     * @return <code>null</code> wenn kein Fault oder keinen faultstring in der Fehlermeldung
     * @throws XPathExpressionException
     */
    public String getFirstFaultDetailElementName() throws XPathExpressionException {
        
    	String name = getXPath().evaluate(
            "name(/soap11:Envelope/soap11:Body/soap11:Fault/soap11:detail/*[1])", 
            getResponseElement());
        
        if (StringUtils.isEmpty(name)) {
            name = getXPath().evaluate(
            	"name(/soap11:Envelope/soap11:Body/soap11:Fault/detail/*[1])",
            	getResponseElement());
        }
        
        if (StringUtils.isEmpty(name)) {
            name = getXPath().evaluate(
                	"name(/soap12:Envelope/soap12:Body/soap12:Fault/soap12:detail/*[1])",
                	getResponseElement());
        }

        if (StringUtils.isEmpty(name)) {
            name = getXPath().evaluate(
                	"name(/soap12:Envelope/soap12:Body/soap12:Fault/detail/*[1])",
                	getResponseElement());
        }

        if (StringUtils.isEmpty(name)) {
            return null;
        }
        return name;
    }

    private XPath getXPath()
    {
        XPath xpath = XPathFactory.newInstance().newXPath();
        xpath.setNamespaceContext(mResponseNamespaceContext);
        return xpath;
    }

    private static void saveClose(InputStream is)
    {
        try
        {
            is.close();
        }
        catch (IOException e)
        {
            // ignore
        }
    }

    /**
     * Konvertiert einen Input Stream in ein DOM Dokument
     * @param response
     * @return das Dokument
     * @throws IOException
     */
    protected static Document parseToDom(InputStream response) throws IOException
    {
        try
        {
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            dbf.setNamespaceAware(true);
            DocumentBuilder builder = dbf.newDocumentBuilder();
            return builder.parse(response);
        }
        catch (SAXException e)
        {
            throw new TechnicalWebserviceCallException("Malformed response: " + e.getMessage(), e);
        }
        catch (ParserConfigurationException e)
        {
            throw new TechnicalWebserviceCallException("Invalid XML-Parser", e);
        }
    }

    /**
     * Request vom Classpath lesen.
     * 
     * @param caller
     *            nicht null
     * 
     * @param name
     *            Name im Classpath (relativ zu caller)
     * @return nicht <code>null</code>
     * @throws RequestTemplateNotFoundException
     */
    private static String readRequest(Class<?> caller, String name)
        throws RequestTemplateNotFoundException
    {
        InputStream is = caller.getResourceAsStream(name);
        if (is == null)
            throw new RequestTemplateNotFoundException(name);

        try
        {
            String result = readStreamToString(is);
            return result;
        }
        catch (IOException e)
        {
            throw new TechnicalWebserviceCallException("Error reading the template " + name, e);
        }
        finally
        {
            saveClose(is);
        }
    }

    /**
     * Einen InputStream in einen String lesen (UTF-8 encodiert).
     * 
     * @param is
     *            nicht <code>null</code>
     * @return nicht <code>null</code>
     * @throws UnsupportedEncodingException
     * @throws IOException
     */
    private static String readStreamToString(InputStream is) throws UnsupportedEncodingException,
        IOException
    {
        Validate.notNull(is, "no input stream");
        InputStreamReader reader = new InputStreamReader(is, "UTF-8");
        StringBuilder b = new StringBuilder();
        char[] buffer = new char[1024];
        int read;
        while ((read = reader.read(buffer)) != -1)
            b.append(buffer, 0, read);
        return b.toString();
    }

	/**
	 * Returns a SSL Socket Factory which does not validate certificate chains
	 * 
	 * @return the SSL Socket Factory
	 */
	private static SSLSocketFactory getTrustAllSocketFactory() {
		if (trustAllSocketFactory == null) {
			try {
				SSLContext sc = SSLContext.getInstance("SSL");
				X509TrustManager[] trustManagers = { new FakeTrustManager() };
				sc.init(null, trustManagers, new SecureRandom());
				trustAllSocketFactory = sc.getSocketFactory();
			} catch (NoSuchAlgorithmException e) {
				throw new RuntimeException("can't initialize SSL Socket Factory", e);
			} catch (KeyManagementException e) {
				throw new RuntimeException("can't initialize SSL Socket Factory", e);
			}
		}
		return trustAllSocketFactory;
	}    
}
