package ch.dvbern.stvbe.security.saml.util;

//	 $Id: NamespaceContextImpl.java,v 1.1 2006/11/22 10:02:42 beph Exp $
//	 Copyright 2004 Sun Microsystems, Inc. All rights reserved.

import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import javax.xml.XMLConstants;
import javax.xml.namespace.NamespaceContext;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * <p>
 * Minimum implementation of NamespaceContext.
 * </p>
 * 
 * <p>
 * Assumes "<em>static</em>" namespace, i.e. NamespaceContext is not updated
 * during parsing. Also assumes "<em>flat</em>" namespace, i.e. only one
 * prefix per Namespace URI.
 * </p>
 */
public class NamespaceContextImpl implements NamespaceContext {

	private static final Log log = LogFactory.getLog(NamespaceContext.class);

	private HashMap<String, String> prefixToNamespaceURI = new HashMap<String, String>();

	private HashMap<String, String> namespaceURIToPrefix = new HashMap<String, String>();

	/**
	 * Bind a prefix to a Namespace URI.
	 * @param prefix		Prefix to bind to <code>namespaceURI</code>.
	 * @param namespaceURI	Namespace URI to bind to <code>prefix</code>.
	 */
	public void bindPrefixToNamespaceURI(String prefix, String namespaceURI) {

		if (log.isTraceEnabled()) {
			log.trace("bindPrefixToNamespaceURI(" + prefix + ", " + namespaceURI + ")");
		}
		
		prefixToNamespaceURI.put(prefix, namespaceURI);
		namespaceURIToPrefix.put(namespaceURI, prefix);
	}

	/**
	 * Get the Namespace URI bound to the prefix.
	 * @param prefix	Lookup Namespace URI that is bound to this prefix.
	 * @return Namespace URI bound to <code>prefix</code>.
	 */
	public String getNamespaceURI(String prefix) {

		if (log.isTraceEnabled()) {
			log.trace("getNamespaceURI(" + prefix + ")");
		}

		if (prefix == null) {
			throw new NullPointerException("prefix may not be null");
		}

		// constants
		if (prefix.equals(XMLConstants.XML_NS_PREFIX)) {
			return XMLConstants.XML_NS_URI;
		}
		if (prefix.equals(XMLConstants.XMLNS_ATTRIBUTE)) {
			return XMLConstants.XMLNS_ATTRIBUTE_NS_URI;
		}

		// default
		if (prefix.equals(XMLConstants.DEFAULT_NS_PREFIX)) {
			if (prefixToNamespaceURI.containsKey(prefix)) {
				return prefixToNamespaceURI.get(prefix);
			} else {
				return XMLConstants.NULL_NS_URI;
			}
		}

		// bound
		if (prefixToNamespaceURI.containsKey(prefix)) {
			return prefixToNamespaceURI.get(prefix);
		}

		// unbound
		return XMLConstants.NULL_NS_URI;
	}

	/**
	 * Get the prefix bound to the Namespace URI.
	 * @param namespaceURI	Lookup prefix that is bound to this Namespace URI.
	 * @return prefix bound to this Namespace URI.
	 */
	public String getPrefix(String namespaceURI) {

		if (log.isTraceEnabled()) {
			log.trace("getPrefix(" + namespaceURI + ")");
		}

		if (namespaceURI == null) {
			throw new NullPointerException("namespaceURI may not be null");
		}

		// constants
		if (namespaceURI.equals(XMLConstants.XML_NS_URI)) {
			return XMLConstants.XML_NS_PREFIX;
		}
		if (namespaceURI.equals(XMLConstants.XMLNS_ATTRIBUTE_NS_URI)) {
			return XMLConstants.XMLNS_ATTRIBUTE;
		}

		// bound
		if (namespaceURIToPrefix.containsKey(namespaceURI)) {
			return namespaceURIToPrefix.get(namespaceURI);
		}

		// mimic "default Namespace URI"
		if (namespaceURI.equals(XMLConstants.NULL_NS_URI)) {
			return XMLConstants.DEFAULT_NS_PREFIX;
		}

		// unbound
		return null;
	}

	/**
	 * Get prefixes bould to Namespace URI.
	 * @param namespaceURI	Lookup prefixes bound to this Namespace URI
	 * @return Iterator over prefixes bound to this Namespace URI.
	 */
	public Iterator<String> getPrefixes(String namespaceURI) {

		if (log.isTraceEnabled()) {
			log.trace("getPrefixes(" + namespaceURI + ")");
		}

		// flat namespace
		String prefix = getPrefix(namespaceURI);
		if (prefix != null) {
			return Collections.singleton(prefix).iterator();
		} else {
			List<String> li = Collections.emptyList();
			return li.iterator();
		}

	}
}
