package ch.dvbern.stvbe.security.saml.verify;

import java.security.PublicKey;
import java.util.Collections;
import java.util.Date;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.opensaml.SAMLAssertion;
import org.opensaml.SAMLException;

/**
 * Standard SAML Assertion Verifier
 * @author GEMA
 */
public class DefaultSamlAssertionVerifier extends AbstractSamlAssertionVerifier {

	private static final Log log = LogFactory.getLog(DefaultSamlAssertionVerifier.class);

	private Set<PublicKey> defaultKeys;
	private Map<String, PublicKey> issuerKeyMapping;
	private boolean signatureRequired = true;
	private boolean validityPeriodRequired = true;
	private boolean validityPeriodIgnored = false; 
	
	/**
	 * Konstruktor
	 */
	public DefaultSamlAssertionVerifier() {
	}

	/**
	 * Konstruktor
	 * @param defaultKeys      die default Public Keys fuer die Signatur Validierung
	 * @param issuerKeyMapping das Mapping zwischen Isssuer URI und Public Keys
	 */
	public DefaultSamlAssertionVerifier(Set<PublicKey> defaultKeys, Map<String, PublicKey> issuerKeyMapping) {
		if ((defaultKeys == null || defaultKeys.isEmpty()) && issuerKeyMapping == null) {
			throw new IllegalArgumentException("either defaultKey or issuerKeyMapping must not be null");
		}
		setDefaultPublicKeys(defaultKeys);
		setIssuerKeyMapping(issuerKeyMapping);
	}

	/**
	 * Legt fest, ob die Gueltigkeitsdauer in der SAML Assertion enthalten
	 * sein muss
	 * @param validityPeriodRequired
	 */
	public void setValidityPeriodRequired(boolean validityPeriodRequired) {
		this.validityPeriodRequired = validityPeriodRequired;
		if (!validityPeriodRequired) {
			log.warn("SAML Assertion Validity Period is NOT ENFORCED");
		}
	}

	/**
	 * Legt fest, ob die SAML Assertion signiert sein muss
	 * @param signatureRequired
	 */
	public void setSignatureRequired(boolean signatureRequired) {
		this.signatureRequired = signatureRequired;
		if (!signatureRequired) {
			log.warn("SAML Assertion Signature Verification DISABLED");
		}
	}
	
	/**
	 * Legt fest, ob der Gltigkeitszeitraum der SAML Assertion ignoriert
	 * werden soll
	 * @param validityPeriodIgnored <code>true</code> falls der Gltigkeitszeitraum ignoriert werden soll
	 */
	public void setValidityPeriodIgnored(boolean validityPeriodIgnored) {
		this.validityPeriodIgnored = validityPeriodIgnored;
		if (!signatureRequired) {
			log.warn("SAML Assertion Validity Period is IGNORED");
		}
	}
	
	/**
	 * Setzt den Default Public Key fuer die Signatur Validierung
	 * @param publicKey den Key
	 */
	public void setDefaultPublicKey(PublicKey publicKey) {
	    if (publicKey == null) {
	        setDefaultPublicKeys(null);
	    } else {
	        setDefaultPublicKeys(Collections.singleton(publicKey));
	    }
	}
	
    /**
     * Setzt die Default Public Keys fuer die Signatur Validierung
     * @param publicKeys die Keys
     */
    public void setDefaultPublicKeys(Set<PublicKey> publicKeys) {
        this.defaultKeys = publicKeys;
    }

    /**
	 * Setut das Mapping zwischen Issuer und Public Key
	 * @param issuerKeyMapping das Mapping
	 */
	public void setIssuerKeyMapping(Map<String, PublicKey> issuerKeyMapping) {
		this.issuerKeyMapping = issuerKeyMapping;
	}
	
	/**
	 * Verifiziert eine SAML Assertion
	 * @param assertion
	 * @throws SamlAssertionVerifyerException
	 */
	public void verify(SAMLAssertion assertion) throws SamlAssertionVerifyerException {
		
		if (assertion == null) {
			if (isAllowUnauthenticated()) {
				log.debug("Allowing unauthenticated access");
				return;
			} else {
				throw new IllegalArgumentException("Login without SAML Assertion is not allowed. SAML Assertion may not be null");
			}
		}

		try {
			
			// Signatur pruefen
			if (signatureRequired) {
				if ((defaultKeys == null || defaultKeys.isEmpty())&& issuerKeyMapping == null) {
					throw new IllegalStateException("no public key set");
				}
				
				// Public Key ermitteln
				Set<PublicKey> pks = defaultKeys;
				if (issuerKeyMapping != null && assertion.getIssuer() != null) {
					// Mapping Verwenden
					log.trace("Using IssuerURI->PublicKey Mapping");
					PublicKey issuerKey = issuerKeyMapping.get(assertion.getIssuer());
					if (issuerKey != null) {
						// Key verwenden, falls gefunden
						pks = Collections.singleton(issuerKey);
						if (log.isDebugEnabled()) {
							log.debug("Public Key found for Issuer: " + assertion.getIssuer());
						}
					} else {
						if (log.isDebugEnabled()) {
							log.debug("No Public Key found for Issuer: " + assertion.getIssuer());
						}
					}
				} else {
					log.trace("Using default PublicKey");
				}
				
				if (log.isDebugEnabled()) {
					log.debug("Using Public Keys: " + pks);
				}
				
				if (pks == null) {
					throw new SamlAssertionVerifyerException("No Public Key found for Issuer: " + assertion.getIssuer());					
				}
				
				// Verifizieren
				boolean verified = false;
                for (Iterator<PublicKey> it = pks.iterator(); it.hasNext() && !verified;) {
                    try {
                        assertion.verify(it.next());
                        verified = true;
                    } catch (SAMLException ex) {
                        log.debug("Verfication failed. Trying next Public Key", ex);
                    }
                }
                
                if (!verified) {
                    throw new SamlAssertionVerifyerException(
                            "SAML Assertion Signature invalid. No Public Key passed the verification");
                }
				
			} else {
				log.warn("SAML Assertion Signature NOT REQUIRED");
				if (assertion.isSigned()) {
					assertion.verify();
				}
			}
			
			// Gueltigkeit pruefen
			if (!validityPeriodIgnored) {
				validateNotBefore(assertion, validityPeriodRequired);
				validateNotOnOrAfter(assertion, validityPeriodRequired);
			} else {
				log.debug("Ignoring Validity 'notOnOrAfter' and 'notBefore' Timestamps")	;		
			}
			
		} catch (SAMLException e) {
			log.info("SAML Assertion verification failed", e);
			throw new SamlAssertionVerifyerException("SAML Assertion verification failed", e);
		}
	}
	
	
	private void validateNotOnOrAfter(SAMLAssertion assertion, boolean required) throws SamlAssertionVerifyerException {
		Date now = getCurrentTime();

		// Gueltig bis extrahieren
		Date until = assertion.getNotOnOrAfter();
		log.debug("SAML Assertion is not valid on or after: " + until);

		// auf null pruefen
		if (until == null) {
			String msg = "SAML Assertion has no 'notOnOrAfter' Timestamp";
			if (required) {
				log.info(msg);
				throw new SamlAssertionVerifyerException(msg);
			} else {
				log.warn(msg);
				return;
			}
		}

		// mit der aktuellen Zeit vergleichen
		if (!now.before(until)) {
			String msg = "SAML Assertion is no longer valid. Valid until: " + until; 
			log.info(msg);
			throw new SamlAssertionVerifyerException(msg);
		}

		log.debug("'notOnOrAfter' Timestamp validated");
	}
	
	private void validateNotBefore(SAMLAssertion assertion, boolean required) throws SamlAssertionVerifyerException {
		Date now = getCurrentTime();

		// Gueltig ab extrahieren
		Date from = assertion.getNotBefore();
		log.debug("SAML Assertion is not valid before: " + from);

		// auf null pruefen
		if (from == null) {
			String msg = "SAML Assertion has no 'notBefore' Timestamp";
			if (required) {
				log.info(msg);
				throw new SamlAssertionVerifyerException(msg);
			} else {
				log.warn(msg);
				return;
			}
		}
		
		// mit der aktuellen Zeit vergleichen
		if (!(now.after(from) || now.equals(from))) {
			String msg = "SAML Assertion is not yet valid. Valid from: "	+ from;
			log.info(msg);
			throw new SamlAssertionVerifyerException(msg);
		}

		log.debug("'notBefore' Timestamp validated");
		
	}
	
	/**
	 * Gbit die aktuelle Uhrzeit zurueck
	 * @return die Uhrzeit
	 */
	protected Date getCurrentTime() {
		return new Date();
	}
	
	/**
	 * Gibt die Default Public Keys zurueck
	 * @return den Key
	 */	
	protected Set<PublicKey> getDefaultKeys() {
		return defaultKeys;
	}
	
	/**
	 * Gibt das Issuer->PublicKey Mapping zurueck
	 * @return das Mapping
	 */
	protected Map<String, PublicKey> getIssuerKeyMapping() {
		return issuerKeyMapping;
	}
	
	/**
	 * Gibt zurueck, ob eine Signatur erforderlich ist
	 * @return <code>true</code> falls eine Signatur erforderlich ist
	 */
	protected boolean isSignatureRequired() {
		return signatureRequired;
	}
	
	/**
	 * Gibt zurueck, ob eine Gueltigkeitsdauer erforderlich ist
	 * @return <code>true</code> falls eine Gueltigkeitsdauer erforderlich ist
	 */
	protected boolean isValidityPeriodRequired() {
		return validityPeriodRequired;
	}
	
	/**
	 * Gibt zurueck, ob eine Gueltigkeitsdauer ignoriert wird
	 * @return <code>true</code> die Gueltigkeitsdauer ignoriert wird
	 */
	protected boolean isValidityPeriodIgnored() {
		return validityPeriodIgnored;
	}
	
}
