package ch.dvbern.stvbe.security.saml.verify;

import java.io.IOException;
import java.security.Principal;
import java.util.Collection;

import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.UnsupportedCallbackException;
import javax.security.auth.login.LoginContext;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.security.SecurityAssociation;
import org.jboss.security.auth.callback.ObjectCallback;
import org.opensaml.SAMLAssertion;

/**
 * Validier eine SAML Asserion mittels JBoss Login Modul
 * @author GEMA
 */
public class JBossJaasSamlAssertionVerifier extends AbstractSamlAssertionVerifier{

	private static final Log log = LogFactory.getLog(JBossJaasSamlAssertionVerifier.class);
	
	private String securityDomain;
	
	/**
	 * Konstruktor
	 */
	public JBossJaasSamlAssertionVerifier() {
	}

	/**
	 * Konstruktor
	 * @param securityDomain der Name der JAAS Domain
	 */
	public JBossJaasSamlAssertionVerifier(String securityDomain) {
		if (securityDomain == null || securityDomain.trim().length() == 0) {
			throw new IllegalArgumentException("security domain name may not be empty");
		}
		setSecurityDomain(securityDomain);
	}
	
	/**
	 * Setzt den Namen der JAAS Domain
	 * @param securityDomain den Namen
	 */
	public void setSecurityDomain(String securityDomain) {
		if (securityDomain == null) {
			throw new IllegalArgumentException("security domain may not be null");
		}
		this.securityDomain = securityDomain;
	}
	
	public void verify(SAMLAssertion assertion) throws SamlAssertionVerifyerException {
		if (assertion == null) {
			if (isAllowUnauthenticated()) {
				log.debug("Performing unauthenticated login");
			} else {
				throw new IllegalArgumentException("Login without SAML Assertion is not allowed. SAML Assertion may not be null");
			}
		}
		
		// Context aufbauen
		try {
	
			SAMLCallbackHandler handler = new SAMLCallbackHandler(assertion);
			LoginContext lc = new LoginContext(securityDomain, handler);
			lc.login();
	
			// Security Context fuer Webapp herstellen
			Collection<Principal> principals = lc.getSubject().getPrincipals();
			if (!principals.isEmpty()) {
				SecurityAssociation.setPrincipal(principals.iterator().next());
			}
			SecurityAssociation.setCredential(assertion);
			SecurityAssociation.setSubject(lc.getSubject());
	
			log.debug("Login successfull");
			
		} catch (Exception e) {
			log.info("Login failed", e);
			throw new SamlAssertionVerifyerException("Login failed", e);
		}
	}
	
	/**
	 * JAAS Callback Handldler fuer eine SAML Assertion
	 * @author GEMA
	 */
	private static class SAMLCallbackHandler implements CallbackHandler {
		private SAMLAssertion samlAssertion;

		/**
		 * Konstruktor
		 * @param samlAssertion die SAML Assertion
		 */
		public SAMLCallbackHandler(SAMLAssertion samlAssertion) {
			this.samlAssertion = samlAssertion;
		}

		public void handle(Callback[] callbacks) throws IOException, UnsupportedCallbackException {
			for (int i = 0; i < callbacks.length; i++) {
				if (callbacks[i] instanceof ObjectCallback) {
					ObjectCallback sc = (ObjectCallback) callbacks[i];
					sc.setCredential(samlAssertion);
				} else {
					throw new UnsupportedCallbackException(callbacks[i], "Unrecognized Callback");
				}
			}
		}
	}
	
	/**
	 * Gibt den Namen der Security Domain zurueck
	 * @return den Namen
	 */
	protected String getSecurityDomain() {
		return securityDomain;
	}

}