package ch.dvbern.stvbe.security.saml.verify;

import java.security.PublicKey;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import ch.dvbern.stvbe.security.saml.key.KeyStoreX509CertificatePublicKeyProvider;
import ch.dvbern.stvbe.security.saml.key.PublicKeyProvider;
import ch.dvbern.stvbe.security.saml.key.ResourceX509CertificatePublicKeyProvider;
import ch.dvbern.stvbe.security.saml.util.ConfigKeys;

/**
 * Factory fuer SAML Assertion Verifiers (Singleton)
 * @author GEMA
 *
 */
public class SamlAssertionVerifierFactory {

	private SamlAssertionVerifierFactory() {
	}
	
	private static final SamlAssertionVerifierFactory instance = new SamlAssertionVerifierFactory();
	
	/**
	 * Gibt die Factory zurueck 
	 * @return die Factory
	 */
	public static SamlAssertionVerifierFactory getInstance() {
		return instance;
	}
	
	/**
	 * Erstellt einen neuen Verifier
	 * @param config die Konfiguration
	 * @return den Verifier
	 */
	public SamlAssertionVerifier createVerifier(Map<String,String> config) {
		Properties props = new Properties();
		props.putAll(config);
		return createVerifier(props);
	}
	
	/**
	 * Erstellt einen neuen Verifier
	 * @param config die Konfiguration
	 * @return den Verifier
	 */
	public SamlAssertionVerifier createVerifier(Properties config) {
		AbstractSamlAssertionVerifier verifier;
		
		String securityDomain = config.getProperty(ConfigKeys.SECURITY_DOMAIN);
		if (securityDomain != null && securityDomain.trim().length() > 0) {
			verifier = new JBossJaasSamlAssertionVerifier(securityDomain);
			
		} else {

			if (Boolean.parseBoolean(config.getProperty(ConfigKeys.SIGNATURE_VERIFICATION_DISABLED))) {
				verifier = new DefaultSamlAssertionVerifier();
				((DefaultSamlAssertionVerifier)verifier).setSignatureRequired(false);
			} else {
				
				// Default Keys aus Zertifikaten ermitteln
				Set<PublicKey> defaultKeys = new HashSet<PublicKey>();
				String cert = config.getProperty(ConfigKeys.CERTIFICATE_RESOURCE_NAME);
				if (cert != null && cert.trim().length() > 0) {
					PublicKeyProvider prov = new ResourceX509CertificatePublicKeyProvider(cert);
					defaultKeys.addAll(prov.getPublicKeys());
				}

                // Default Keys aus Keystore ermitteln
				String keystorePath = config.getProperty(ConfigKeys.TRUSTSTORE_PATH);
                if (keystorePath != null && keystorePath.trim().length() > 0) {
                    String keystorePassword = config.getProperty(ConfigKeys.TRUSTSTORE_PASSWORD);
                    PublicKeyProvider prov = new KeyStoreX509CertificatePublicKeyProvider(keystorePath, keystorePassword);
                    defaultKeys.addAll(prov.getPublicKeys());
                }
				
				// Mapping ermitteln
				Map<String, PublicKey> issuerKeyMapping = parseIssuerKeyMapping(config);
				
				if (cert == null && keystorePath == null && issuerKeyMapping.isEmpty() ) {
					throw new IllegalArgumentException(ConfigKeys.CERTIFICATE_RESOURCE_NAME + 
					        " or " + ConfigKeys.TRUSTSTORE_PATH + 
							" must be set if " + ConfigKeys.SECURITY_DOMAIN + " is empty");
				}
				
				verifier = new DefaultSamlAssertionVerifier(defaultKeys, issuerKeyMapping);
			}

			if (Boolean.parseBoolean(config.getProperty(ConfigKeys.TIME_VALIDITY_OPTIONAL))) {
				((DefaultSamlAssertionVerifier)verifier).setValidityPeriodRequired(false);
			}

			if (Boolean.parseBoolean(config.getProperty(ConfigKeys.TIME_VALIDITY_IGNORE))) {
				((DefaultSamlAssertionVerifier)verifier).setValidityPeriodIgnored(true);
			}
		}
		
		// allow unauthenticated verarbeiten
		boolean allowUnauthenticated = Boolean.parseBoolean(config.getProperty(ConfigKeys.ALLOW_UNAUTHENTICATED));
		verifier.setAllowUnauthenticated(allowUnauthenticated);
		
		return verifier;
	}
	
	/**
	 * Erstellt das Mapping zwischen Issuern und Public Keys
	 * @param config die Konfiguration
	 * @return das Mapping
	 */
	private Map<String, PublicKey> parseIssuerKeyMapping(Properties config) {
		ResourceX509CertificatePublicKeyProvider prov = new ResourceX509CertificatePublicKeyProvider();
		Map<String, PublicKey> mapping = new HashMap<String, PublicKey>();
		
		Set<Map.Entry<Object,Object>> entries = config.entrySet();
		for (Map.Entry<Object, Object> entry : entries) {
			String keyName = entry.getKey().toString();
			if (!keyName.startsWith(ConfigKeys.ASSERTION_ISSUER + ".")) {
				// Ignorieren, falls kein Issuer
				continue;
			}
			
			// laenge pruefen
			if (keyName.length() <= ConfigKeys.ASSERTION_ISSUER.length() + 1) {
				throw new IllegalArgumentException(keyName + " must be followed by an id");
			}
			
			// index parsen
			String id = keyName.substring(ConfigKeys.ASSERTION_ISSUER.length() + 1);

			// Issuer ermitteln
			String issuer = (String)entry.getValue();
			
			// Certifcate Resource ermitteln
			String resourceKey = ConfigKeys.CERTIFICATE_RESOURCE_NAME + "." + id;
			String cert = config.getProperty(resourceKey);
			if (cert == null || cert.trim().length() == 0) {
				throw new IllegalArgumentException("Property for certificate resource not found:" + resourceKey );
			}

			// Public Key laden und im Mapping ablegen
			prov.setResourceName(cert);
			mapping.put(issuer, prov.getPublicKeys().iterator().next());
			
		}
		
		return mapping;
		
	}
	
	
	
	
	
}
