package ch.dvbern.stvbe.security.saml.wss4j;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.Date;
import java.util.Iterator;
import java.util.Properties;

import javax.security.auth.Subject;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.ws.security.components.crypto.Crypto;
import org.apache.ws.security.saml.SAMLIssuer;
import org.jboss.security.SecurityAssociation;
import org.opensaml.SAMLAssertion;
import org.w3c.dom.Document;

import ch.dvbern.stvbe.security.saml.issuer.IssuedToken;
import ch.dvbern.stvbe.security.saml.issuer.IssuerException;
import ch.dvbern.stvbe.security.saml.issuer.renew.RenewService;
import ch.dvbern.stvbe.security.saml.issuer.webservice.ApplicationInfo;
import ch.dvbern.stvbe.security.saml.issuer.webservice.DefaultApplicationInfo;
import ch.dvbern.stvbe.security.saml.issuer.webservice.STSRenewerService;
import ch.dvbern.stvbe.security.saml.util.StringUtils;

/**
 * SAML Issuer welcher eine bestehende SAML Assertion aus dem  
 * @author GEMA
 */
public class PassthroughSamlIssuer implements SAMLIssuer {

	private static final Log log = LogFactory.getLog(PassthroughSamlIssuer.class);
    private static final long EXPIRATION_MARGIN = 300000; // 5 Min
    private static final String RENEWER_SERVICE_URL_PROPERTY = "sts.renewer.service.url";
    private static final String SSL_CERTIFICATE_VALIDATION_DISABLED_PROPERTY = "ssl.certificate.validation.disabled";
    
    private STSRenewerService renewer;
    private ApplicationInfo appInfo;
    
	/**
	 * Konstruktor
	 */
	public PassthroughSamlIssuer() {
    	this (new Properties());
	}

	/**
	 * Konstruktor
	 * @param prop
	 */
	public PassthroughSamlIssuer(Properties prop) {
		// App Info
		appInfo = new DefaultApplicationInfo();
		
		// Renewer
		String urlValue = getSamlAssertionRenewerUrl(prop);
		if (StringUtils.isEmpty(urlValue)) {
			log.info("Keine SAML Assertion Renewer Adresse konfiguriert");
			renewer = null;
		} else {
			try {
				URI renewerServiceUri = new URI(urlValue);
				renewer = new STSRenewerService(renewerServiceUri, appInfo);
				renewer.setSslCertifcateValidationDisabled(isSslCertificateValidationDisabled(prop));
				if (renewer.isSslCertifcateValidationDisabled()) {
					log.warn("Die Validierung der SSL Zertifikate ist ausgeschaltet");
				}
			} catch (URISyntaxException e) {
				throw new RuntimeException("SAML Assertion Renewer Adresse ungueltig", e);
			}
		}
	}
	
	public Crypto getIssuerCrypto() {
		return null;
	}

	public String getIssuerKeyName() {
		return null;
	}

	public String getIssuerKeyPassword() {
		return null;
	}

	public boolean isSenderVouches() {
		return false;
	}

	public SAMLAssertion newAssertion() {
		
		Object credential = getSamlCredential();

		SAMLAssertion assertion;
		if (credential instanceof SAMLAssertion) {
			// normale SAML Assertion
			assertion = (SAMLAssertion)credential;
		} else if (credential instanceof IssuedToken) {
			
			// Pruefen, ob der Renewer konfiguriert ist
			if (getSamlAssertionRenewer() == null) {
				throw new RuntimeException("Die Adresse des SAML Assertion Renewer Service " +
						"muss konfiguriert werden, falls ein erneuerbares Token verwendet wird");
			}
			
			// Issued Token (mit renewal Funktionalitaet)
			IssuedToken token = (IssuedToken)credential;
			if (isTokenExpired(token)) {
				// Token erneuern, falls abgelaufen 
				token = renewToken(token);
				setSamlCredential(token);
			} 
			assertion = token.getAssertion();
		} else {
			log.debug("credential is not a SAML Assertion or a Issued Token");
			throw new RuntimeException("credential is not a SAML Assertion or a Issued Token: " + credential.getClass());
		}
		
		if (log.isDebugEnabled()) {
			log.debug("reusing SAML Assertion:\n" + assertion);
		}
		
		return assertion;
	}
	
	/**
	 * Ermittelt das SAML Credential aus dem JBoss Security Context
	 * @return das Credential
	 */
	private Object getSamlCredential() {
		Object credential = SecurityAssociation.getCredential();
		if (credential == null) {
			log.debug("no credential found. probably not logged-in");
			throw new RuntimeException("no credential found. probably not logged-in");
		}
		
		// Credential ist bereits die SAML Assertion oder das IssuedToken
		if (credential instanceof SAMLAssertion || credential instanceof IssuedToken) {
			return credential;
		}
		
		// Credential wurde vom Login Module in den Private Credentials des Subjects ab gelegt
		Subject subject = SecurityAssociation.getSubject();
		if (subject  != null && subject.getPrivateCredentials() != null) {
			for (Iterator<Object> it = subject.getPrivateCredentials().iterator(); it.hasNext();) {
				credential = it.next();
				if (credential instanceof SAMLAssertion || credential instanceof IssuedToken) {
					return credential;
				}
			}
		}

		log.debug("No SAML Assertion or Issued Token Credential found");
		throw new RuntimeException("No SAML Assertion or Issued Token Credential found");
		
	}
	
	private void setSamlCredential(IssuedToken newToken) {
		// Falls das Credential ein IssuedToken ist, wird dieses ersetzt
		if (SecurityAssociation.getCredential() instanceof IssuedToken) {
			SecurityAssociation.setCredential(newToken);	
		}
		
		// IssuedToken in den private Credentials des Subjects ersetzten
		Subject subject = SecurityAssociation.getSubject();
		if (subject  != null && subject.getPrivateCredentials() != null) {
			// alle IssuedToken entfernen
			for(Iterator<Object> it = subject.getPrivateCredentials().iterator(); it.hasNext();) {
				if (it.next() instanceof IssuedToken) {
					it.remove();
				}
			}
			
			// neues Token setzen
			subject.getPrivateCredentials().add(newToken);
		}
		
	}

	public void setInstanceDoc(Document arg0) {
	}

	public void setUserCrypto(Crypto arg0) {
	}

	public void setUsername(String arg0) {
	}

	/**
	 * Gibt zurueck, ob eine SAML Assertion fuer den aktuellen Benutzer vorhanden ist 
	 * @return <code>true</code> falls eine Assertion vorhanden ist
	 */
	public boolean isAssertionAvailable() {
		return SecurityAssociation.getCredential() != null;
	}
	

	/**
	 * Erneutert das uebergebene Toekn
	 * @param oldToken das alte Token
	 * @return das neue Token
	 */
	protected IssuedToken renewToken(IssuedToken oldToken) {
        try {
			return getSamlAssertionRenewer().getToken(oldToken);
		} catch (IssuerException e) {
			throw new RuntimeException("SAML Assertion kann nicht erneuert werden", e);
		}
	}
	
    /**
     * Testet, ob das aktuelle Token abgelaufen ist oder in Kuerze ablaeuft.
     * @param token das Toeken
     * @return <code>true</code> wenn Token bald ablaeuft
     */
    protected boolean isTokenExpired(IssuedToken token) {
        Date expiration = token.getAssertion().getNotOnOrAfter();
        if (expiration == null) {
            return false;
        }
        return new Date().getTime() > expiration.getTime() - EXPIRATION_MARGIN;
    }
    
    /**
     * Gibt den Service zum Erneuern der Tokens zurueck
     * @return den Service
     */
    protected RenewService getSamlAssertionRenewer() {
    	return renewer;
    }
    
    /**
     * Gibt die Adresse des Service zum Erneuern der Tokens zurueck
     * @param config die Konfiguration
     * @return die Adresse
     */
    protected String getSamlAssertionRenewerUrl(Properties config) {
		return config.getProperty(RENEWER_SERVICE_URL_PROPERTY);
    }

    /**
     * Gibt zurueck, ob die Validierung der SSL Zertifikate ausgeschaltet ist
     * @param config die Konfiguration
     * @return <code>true</code> falls die Validierung ausgeschaltet ist
     */
    protected boolean isSslCertificateValidationDisabled(Properties config) {
		String value = config.getProperty(SSL_CERTIFICATE_VALIDATION_DISABLED_PROPERTY);
		if (StringUtils.isEmpty(value)) {
			return false;
		} else {
			return Boolean.valueOf(value);
		}
    }
}
