package ch.dvbern.stvbe.security.saml.wss4j;

import java.security.cert.X509Certificate;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Collections;
import java.util.Date;
import java.util.Properties;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.ws.security.WSSecurityException;
import org.apache.ws.security.components.crypto.Crypto;
import org.apache.ws.security.components.crypto.CryptoFactory;
import org.apache.ws.security.saml.SAMLIssuer;
import org.apache.xml.security.exceptions.XMLSecurityException;
import org.apache.xml.security.keys.KeyInfo;
import org.apache.xml.security.keys.content.X509Data;
import org.apache.xml.security.signature.XMLSignature;
import org.opensaml.SAMLAssertion;
import org.opensaml.SAMLAuthenticationStatement;
import org.opensaml.SAMLException;
import org.opensaml.SAMLNameIdentifier;
import org.opensaml.SAMLStatement;
import org.opensaml.SAMLSubject;
import org.w3c.dom.Document;

/**
 * Einfacher SAML Issuer, welcher eine SAML Assertion aufgrund der Konfiguration in einem
 * Properties File erstellt.
 * @author GEMA
 */
public class SimpleSamlIssuer implements SAMLIssuer {
	private static Log log = LogFactory.getLog(SimpleSamlIssuer.class);

	private Properties properties = null;

	private Crypto issuerCrypto = null;

	private String issuerKeyPassword = null;

	private String issuerKeyName = null;

	private boolean senderVouches = true;

	private String[] confirmationMethods = new String[1];

	private Crypto userCrypto = null;

	private String username = null;
	
	private boolean signAssertion = true;
	
	private boolean addTimeContstraints = true;
	
	private Document instanceDoc;
	
	private long validityMs = 600000; 
	

	/**
	 * Konstruktor
	 */
	public SimpleSamlIssuer() {
	}

	/**
	 * Konstruktor
	 * @param prop
	 */
	public SimpleSamlIssuer(Properties prop) {
		/*
		 * if no properties .. just return an instance, the rest will be done
		 * later or this instance is just used to handle certificate conversions
		 * in this implementatio
		 */
		if (prop == null) {
			return;
		}
		properties = prop;

		String cryptoProp = properties
				.getProperty("org.apache.ws.security.saml.issuer.cryptoProp.file");
		if (cryptoProp != null) {
			issuerCrypto = CryptoFactory.getInstance(cryptoProp);
			issuerKeyName = properties
					.getProperty("org.apache.ws.security.saml.issuer.key.name");
			issuerKeyPassword = properties
					.getProperty("org.apache.ws.security.saml.issuer.key.password");
		}

		if ("senderVouches".equals(properties
				.getProperty("org.apache.ws.security.saml.confirmationMethod"))) {
			confirmationMethods[0] = SAMLSubject.CONF_SENDER_VOUCHES;
		} else if ("keyHolder".equals(properties
				.getProperty("org.apache.ws.security.saml.confirmationMethod"))) {
			confirmationMethods[0] = SAMLSubject.CONF_HOLDER_KEY;
			senderVouches = false;
		} else {
			// throw something here - this is a mandatory property
		}
	}

	/**
	 * Creates a new <code>SAMLAssertion</code>. <p/> <p/> A complete
	 * <code>SAMLAssertion</code> is constructed.
	 * 
	 * @return SAMLAssertion
	 */
	public SAMLAssertion newAssertion() { // throws Exception {
		log.debug("Begin add SAMLAssertion token...");

		/*
		 * if (senderVouches == false && userCrypto == null) { throw
		 * exception("need user crypto data to insert key") }
		 */
		// Issuer must enable crypto fubctions to get the issuer's certificate
		String issuer = getIssuerName();
		String name = getSubjectName();
		String qualifier = getSubjectNameQualifier();
		try {
			SAMLNameIdentifier nameId = new SAMLNameIdentifier(name, qualifier,
					"");
			String subjectIP = null;
			String authMethod = null;
			if ("password"
					.equals(properties
							.getProperty("org.apache.ws.security.saml.authenticationMethod"))) {
				authMethod = SAMLAuthenticationStatement.AuthenticationMethod_Password;
			}
			Date authInstant = new Date();

			SAMLSubject subject = new SAMLSubject(nameId, Arrays
					.asList(confirmationMethods), null, null);
			SAMLStatement[] statements = { new SAMLAuthenticationStatement(
					subject, authMethod, authInstant, subjectIP, null, null) };
			
			Date notBefore = null;
			Date notOnOrAfter = null;
			
			if (addTimeContstraints) {
				Calendar cal = Calendar.getInstance();
				cal.add(Calendar.MINUTE, -5);
				notBefore = cal.getTime();
				
				long time = System.currentTimeMillis();
				time += validityMs;
				notOnOrAfter = new Date(time);
			}
			
			SAMLAssertion sa = new SAMLAssertion(issuer, notBefore , notOnOrAfter, null, null, Arrays
					.asList(statements));
					

			if (!senderVouches) {
				KeyInfo ki = new KeyInfo(instanceDoc);
				try {
					X509Certificate[] certs = userCrypto.getCertificates(username);
					X509Data certElem = new X509Data(instanceDoc);
					certElem.addCertificate(certs[0]);
					ki.add(certElem);
				} catch (WSSecurityException e) {
					return null;
				} catch (XMLSecurityException e) {
					return null;
				}
				subject.setKeyInfo(ki);
				sa = signAssertion(sa, subject, true);
			} else {
				sa = signAssertion(sa, subject, true);
			}
			return sa;
		} catch (SAMLException ex) {
			ex.printStackTrace();
			throw new RuntimeException(ex.toString());
		}
	}
	
	private SAMLAssertion signAssertion(SAMLAssertion sa, SAMLSubject subject, boolean addCerts) throws SAMLException{
		if (!signAssertion) {
			return sa;
		}
		
		// prepare to sign the SAML token
		try {
			X509Certificate[] issuerCerts = issuerCrypto
					.getCertificates(issuerKeyName);

			String sigAlgo = XMLSignature.ALGO_ID_SIGNATURE_RSA;
			String pubKeyAlgo = issuerCerts[0].getPublicKey()
					.getAlgorithm();
			log.debug("automatic sig algo detection: " + pubKeyAlgo);
			if (pubKeyAlgo.equalsIgnoreCase("DSA")) {
				sigAlgo = XMLSignature.ALGO_ID_SIGNATURE_DSA;
			}
			java.security.Key issuerPK = issuerCrypto.getPrivateKey(
					issuerKeyName, issuerKeyPassword);
			
			sa.sign(sigAlgo, issuerPK, addCerts ? Arrays.asList(issuerCerts) : Collections.EMPTY_LIST);
			
			
			// Hack fuer WLS Bug 
//			Node node = sa.toDOM();
//			removeNewlines(node.getOwnerDocument().getElementsByTagNameNS("http://www.w3.org/2000/09/xmldsig#", "SignatureValue"));
//			removeNewlines(node.getOwnerDocument().getElementsByTagNameNS("http://www.w3.org/2000/09/xmldsig#", "KeyInfo"));
			
			/*
			if (nl.getLength() > 0) {
				Node n = nl.item(0).getNextSibling();
				if (n.getNodeType() == Node.TEXT_NODE) {
					if (n.getTextContent().equals("\n")) {
						n.getParentNode().removeChild(n);
					}
				}
			}
			*/
			
//			sa = new SAMLAssertion((Element)node);
			
			
		} catch (WSSecurityException e1) {
			e1.printStackTrace();
			return null;
		} catch (Exception e1) {
			e1.printStackTrace();
			return null;
		}
		
		return sa;
		
	}
	
	/**
	 * @param userCrypto
	 *            The userCrypto to set.
	 */
	public void setUserCrypto(Crypto userCrypto) {
		this.userCrypto = userCrypto;
	}

	/**
	 * @param username
	 *            The username to set.
	 */
	public void setUsername(String username) {
		this.username = username;
	}

	/**
	 * @return Returns the issuerCrypto.
	 */
	public Crypto getIssuerCrypto() {
		return issuerCrypto;
	}

	/**
	 * @return Returns the issuerKeyName.
	 */
	public String getIssuerKeyName() {
		return issuerKeyName;
	}

	/**
	 * @return Returns the issuerKeyPassword.
	 */
	public String getIssuerKeyPassword() {
		return issuerKeyPassword;
	}

	/**
	 * @return Returns the senderVouches.
	 */
	public boolean isSenderVouches() {
		return senderVouches;
	}

	/**
	 * @param instanceDoc
	 *            The instanceDoc to set.
	 */
	public void setInstanceDoc(Document instanceDoc) {
		this.instanceDoc = instanceDoc;
	}
	
	/**
	 * @param signAssertion
	 */
	public void setSignAssertion(boolean signAssertion) {
		this.signAssertion = signAssertion;
	}

	/**
	 * @param validityMs
	 */
	public void setValidityMs(long validityMs) {
		this.validityMs = validityMs;
	}

	/**
	 * @param addTimeContstraints
	 */
	public void setAddTimeContstraints(boolean addTimeContstraints) {
		this.addTimeContstraints = addTimeContstraints;
	}

	/**
	 * Gibt den Namen des Issuers zurueck 
	 * @return den Namen
	 */
	protected String getIssuerName() {
		return properties.getProperty("org.apache.ws.security.saml.issuer");
	}
	
	/**
	 * Gibt den Namen des Subjekts (Benutzername in der SAML Assertion) zurueck 
	 * @return den Namen
	 */
	protected String getSubjectName() {
		return properties.getProperty("org.apache.ws.security.saml.subjectNameId.name");
	}
	
	/**
	 * Gibt den Qualifier fuer den Namen des Subjekts (Benutzername in der SAML Assertion) zurueck 
	 * @return den Namen
	 */
	protected String getSubjectNameQualifier() {
		return properties.getProperty("org.apache.ws.security.saml.subjectNameId.qualifier");
	}

}
